/*-
 * Copyright (c) 2012
 * 	Swinburne University of Technology, Melbourne, Australia.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * A PAM session module which can be used to manage filesystem quotas.
 * XXX: Currently only handles ZFS datasets.
 * XXX: Does not yet support groups.
 *
 * Author: Lawrence Stewart <lastewart at swin dot edu dot au>
 */

#include <sys/types.h>
#include <sys/queue.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	PAM_SM_SESSION
#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <security/pam_mod_misc.h>

/* ZFS includes from ${SRCDIR}/cddl and ${SRCDIR}/sys/cddl */
#include <libzfs.h>

struct peruserfsquotaconf {
	char *user;
	int64_t quota;
	SLIST_ENTRY(peruserfsquotaconf) next;
};

struct fsquotaconf {
	char *fspath;
	int64_t default_quota;
	SLIST_HEAD(fsuserquotas, peruserfsquotaconf) usersconf;
	SLIST_ENTRY(fsquotaconf) next;
};

/*
 * Setting to non-zero causes PAM_LOG_DEBUG messages to be logged to syslog with
 * LOG_DEBUG priority. See openpam_log(3) for more information.
 */
extern int _openpam_debug;

/* Username the session is being established for. */
const char *session_user;

SLIST_HEAD(fsquotaconfhead, fsquotaconf) fsquotaconfs =
    SLIST_HEAD_INITIALIZER(fsquotaconfs);

static int
ltrim(char **str)
{
	int n;

	n = 0;

	/* Remove leading whitespace. */
	while (isspace(**str)) {
		n++;
		(*str)++;
	}

	return (n);
}

static int
rtrim(char **str)
{
	int i, n;

	n = 0;

	/* Remove trailing whitespace. */
	i = strlen(*str) - 1;
	while (i >= 0 && isspace((*str)[i])) {
		n++;
		(*str)[i--] = '\0';
	}

	return (n);
}

static int
trim(char **str)
{

	return (rtrim(str) + ltrim(str));
}

/*
 * Attempt to set property to value for given dataset.
 * Ideas cribbed from:
 *   zfs_main.c:main()
 *   zfs_main.c:set_callback()
 *   zfs_iter.c:zfs_for_each()
 * Return 0 on success, -1 on libzfs init failure or >0 (libzfs errno).
 */
static int
zfs_set_prop(const char *propname, const char *propval, const char *dataset)
{
	libzfs_handle_t *lzhp = NULL;
	zfs_handle_t *zhp = NULL;
	int err = 0;

	lzhp = libzfs_init();
	if (lzhp == NULL) {
		err = -1;
		goto cleanup;
	}

	zhp = zfs_open(lzhp, dataset, ZFS_TYPE_DATASET);
	if (zhp == NULL) {
		err = libzfs_errno(lzhp);
		openpam_log(PAM_LOG_ERROR, "%s - returned error %d (%s) "
		    "trying to open ZFS dataset \"%s\"", __FILE__, err,
		    libzfs_error_description(lzhp), dataset);
		goto cleanup;
	}

	if (zfs_prop_set(zhp, propname, propval) != 0) {
		err = libzfs_errno(lzhp);
		openpam_log(PAM_LOG_ERROR, "%s - returned error %d (%s) "
		    "trying to set ZFS property \"%s=%s\" on dataset \"%s\"",
		    __FILE__, err, libzfs_error_description(lzhp), propname,
		    propval, dataset);
	}

cleanup:
	if (zhp != NULL)
		zfs_close(zhp);

	if (lzhp != NULL)
		libzfs_fini(lzhp);

	return (err);
}

static struct fsquotaconf *
get_fsquotaconf(const char *fspath)
{
	struct fsquotaconf *tmpquotaconf;
	char *fspath2;

	/* Check if an appropriate fsquotaconf struct exists yet. */
	SLIST_FOREACH(tmpquotaconf, &fsquotaconfs, next) {
		if (strcmp(tmpquotaconf->fspath, fspath) == 0)
			break;
	}

	if (tmpquotaconf == NULL) {
		/* No fsquotaconf struct exists yet for this filesystem. */
		tmpquotaconf = malloc(sizeof(struct fsquotaconf));
		fspath2 = strdup(fspath);
		if (tmpquotaconf != NULL && fspath2 != NULL) {
			tmpquotaconf->fspath = fspath2;
			tmpquotaconf->default_quota = -1;
			SLIST_INIT(&tmpquotaconf->usersconf);
			SLIST_INSERT_HEAD(&fsquotaconfs, tmpquotaconf, next);
			PAM_LOG("%s - inserting fsquotaconf %p for fspath %s",
			__FILE__, tmpquotaconf, fspath);
		} else {
			free(tmpquotaconf);
			free(fspath2);
		}
	}

	return (tmpquotaconf);
}

static struct peruserfsquotaconf *
get_peruserfsquotaconf(struct fsquotaconf *fsqc, const char *user)
{
	struct peruserfsquotaconf *tmpuserquotaconf;
	char *user2;

	/* Check if an appropriate peruserfsquotaconf struct exists yet. */
	SLIST_FOREACH(tmpuserquotaconf, &fsqc->usersconf, next) {
		if (strcmp(tmpuserquotaconf->user, user) == 0)
			break;
	}

	if (tmpuserquotaconf == NULL) {
		/* No peruserfsquotaconf struct exists for the filesystem. */
		tmpuserquotaconf =
		    malloc(sizeof(struct peruserfsquotaconf));
		user2 = strdup(user);
		if (tmpuserquotaconf != NULL && user2 != NULL) {
			tmpuserquotaconf->user = user2;
			tmpuserquotaconf->quota = -1;
			SLIST_INSERT_HEAD(&fsqc->usersconf, tmpuserquotaconf,
			    next);
			PAM_LOG("%s - inserting peruserfsquotaconf %p for "
			    "fspath %s user %s", __FILE__, tmpuserquotaconf,
			    fsqc->fspath, user);
		}
	}

	return (tmpuserquotaconf);
}

static int
parse_section_line(const char *section, int *skipsection)
{
	int err;

	err = 0;
	*skipsection = 0;
        /* XXX: Verify section name corresponds to a valid dataset. */

	return (err);
}

static int
parse_prop_line(const char *section, const char *key, const char *val)
{
	struct fsquotaconf *tmpquotaconf;
	struct peruserfsquotaconf *tmpuserquotaconf;
	char *key2, *keyname, *user;
	int64_t *quotavar;
	int err;

	err = 0;

	key2 = strdup(key);
	if (key2 == NULL) {
		err = ENOMEM;
		goto end;
	}

	keyname = key2;
	user = strchr(key2, ':');
	if (user != NULL) {
		/* Parsing a per-user key/value pair e.g. "quota:user = 1G". */
		*user = '\0'; /* NULL out the ':'. */
		user++; /* Point user at char after ':'. */
	}

	PAM_LOG("%s - section=\"%s\", key=\"%s\", keyname=\"%s\", val=\"%s\", "
	    "user=\"%s\"", __FILE__, section, key, keyname, val, user);

	tmpquotaconf = get_fsquotaconf(section);
	if (tmpquotaconf == NULL) {
		err = ENOMEM;
		goto end;
	}

	if (user != NULL) {
		tmpuserquotaconf = get_peruserfsquotaconf(tmpquotaconf, user);
		if (tmpuserquotaconf == NULL) {
			err = ENOMEM;
			goto end;
		}
		quotavar = &tmpuserquotaconf->quota;
	} else {
		quotavar = &tmpquotaconf->default_quota;
	}

	if (strcmp("quota", keyname) == 0) {
		if (strcmp("none", val) == 0)
			*quotavar = -1;
		else {
			*quotavar = strtoull(val, NULL, 10);
			switch (val[strlen(val) - 1]) {
			case 'k':
			case 'K':
				*quotavar *= 1024;
				break;
			case 'm':
			case 'M':
				*quotavar *= 1048576;
				break;
			case 'g':
			case 'G':
				*quotavar *= 1073741824;
				break;
			default:
				break;
			}

			/* Treat quota=0 the same as quota=none. */
			if (*quotavar == 0)
				*quotavar = -1;
		}
	}

end:
	free(key2);

	return (err);
}

static int
parse_config_file(const char *cfpath)
{
	FILE *f;
	char cursection[1024], line[1024];
	char *key, *p, *val;
	int err, line_no, npropsthissection, nsections, skipsection;

	err = line_no = npropsthissection = nsections = skipsection = 0;

	PAM_LOG("%s - attempting to parse config file at \"%s\"", __FILE__,
	    cfpath);

	f = fopen(cfpath, "r");
	if (f == NULL) {
		err = errno;
		openpam_log(PAM_LOG_ERROR, "%s - can't load config file %s",
		    __FILE__, cfpath);
		err = ENOENT;
		goto err;
	}

	while (fgets(line, sizeof(line), f)) {
		line_no++;
		p = line;

		trim(&p);

		/* Skip comments and whitespace-only lines. */
		if (!strlen(p) || !sscanf(p, "%[^\r\n]\n", p) || *p == '#' ||
		    *p == ';')
			continue;

		if (*p == '[') {
			/* Parse section name e.g. "[blah]". */
			p += 1; /* Make line point to char after '['. */

			if (p[strlen(p) - 1] != ']') {
				openpam_log(PAM_LOG_ERROR,
				    "%s - invalid section format \"%s\"",
				    __FILE__, p);
				err = EFTYPE;
				goto err;
			}

			if (npropsthissection == 0 && nsections > 0) {
				openpam_log(PAM_LOG_ERROR,
				    "%s - no config in section \"%s\"",
				    __FILE__, cursection);
				err = EFTYPE;
				goto err;
			}

			p[strlen(p) - 1] = '\0'; /* Remove trailing ']'. */
			trim(&p);
			strlcpy(cursection, p, sizeof(cursection));
			npropsthissection = 0;
			if (parse_section_line(cursection, &skipsection) != 0) {
				openpam_log(PAM_LOG_ERROR,
				    "%s - failed to parse section \"%s\"",
				    __FILE__, cursection);
				err = EFTYPE;
				goto err;
			}
			nsections++;
		} else if (skipsection) {
			continue;
		} else {
			/* Parsing a "property" line e.g. "key = val". */
			key = val = p;
			strsep(&val, "=");
			if (val == NULL) {
				openpam_log(PAM_LOG_ERROR,
				    "%s - config line %d is invalid",
				    __FILE__, line_no);
				err = EFTYPE;
				goto err;
			}

			/*
			 * Remove trailing whitespace from key and leading
			 * whitespace from val.
			 */
			rtrim(&key);
			ltrim(&val);

			if (strlen(key) == 0 || strlen(val) == 0) {
				openpam_log(PAM_LOG_ERROR,
				    "%s - config line %d is invalid",
				    __FILE__, line_no);
				err = EFTYPE;
				goto err;
			}

			if (parse_prop_line(cursection, key, val) != 0) {
				openpam_log(PAM_LOG_ERROR,
				    "%s - config line %d is invalid",
				    __FILE__, line_no);
				err = EFTYPE;
				goto err;
			}

			npropsthissection++;
		}
	}

err:
	if (f != NULL)
		fclose(f);

	return (err);
}

static void
cleanup(void)
{
	struct fsquotaconf *tmpquotaconf;
	struct peruserfsquotaconf *tmpuserquotaconf;

	while (!SLIST_EMPTY(&fsquotaconfs)) {
		tmpquotaconf = SLIST_FIRST(&fsquotaconfs);
		SLIST_REMOVE_HEAD(&fsquotaconfs, next);

		PAM_LOG("%s - cleanup struct fsquotaconf %p (fspath=\"%s\", "
		    "default_quota=%jd)", __FILE__, tmpquotaconf,
		    tmpquotaconf->fspath, tmpquotaconf->default_quota);

		while (!SLIST_EMPTY(&tmpquotaconf->usersconf)) {
			tmpuserquotaconf =
			    SLIST_FIRST(&tmpquotaconf->usersconf);
			SLIST_REMOVE_HEAD(&tmpquotaconf->usersconf, next);

			PAM_LOG("%s - cleanup struct tmpuserquotaconf %p "
			    "attached to struct fsquotaconf %p "
			    "(user=\"%s\", quota=%jd)", __FILE__,
			    tmpuserquotaconf, tmpquotaconf,
			    tmpuserquotaconf->user, tmpuserquotaconf->quota);

			free(tmpuserquotaconf->user);
			free(tmpuserquotaconf);
		}

		free(tmpquotaconf->fspath);
		free(tmpquotaconf);
	}
}

static int
no_default_for_user(const char *user_to_check, const char *dflt_exempt_users)
{
	char *next_in_list, *user;
	int skip_default = 0;

	if (dflt_exempt_users == NULL || strlen(dflt_exempt_users) == 0)
		return (skip_default);

	next_in_list = strdup(dflt_exempt_users);

	do {
		user = next_in_list;
		next_in_list = strchr(next_in_list, ',');
		if (next_in_list != NULL) {
			*next_in_list = '\0';
			next_in_list++;
		}

		if (strlen(user_to_check) == strlen(user)
		    && strcmp(user_to_check, user) == 0)
			skip_default = 1;
	} while (next_in_list != NULL && !skip_default);

	free(next_in_list);

	return (skip_default);
}

/*
 * Set filesystem quotas for the user.
 * XXX: Only handles ZFS datasets currently.
 */
static int
set_quotas(const char *dflt_exempt_users)
{
	struct fsquotaconf *tmpquotaconf;
	struct peruserfsquotaconf *tmpuserquotaconf;
	char quotaprop[64] = "userquota@";
	char quotaval[64] = "none";
	int64_t *quotavar;
	int err, quota_divisor;
	char suffix;

	err = 0;
	strlcat(quotaprop, session_user, sizeof(quotaprop));

	/* For each filesystem path specified in the config file. */
	SLIST_FOREACH(tmpquotaconf, &fsquotaconfs, next) {
		SLIST_FOREACH(tmpuserquotaconf, &tmpquotaconf->usersconf, next) {
			if (strcmp(session_user, tmpuserquotaconf->user) == 0)
				break;
		}

		/*
		 * If there is NOT a specific override quota configuration for
		 * this user, use the default for the filesystem.
		 */
		if (tmpuserquotaconf == NULL) {
			if (no_default_for_user(session_user,
			    dflt_exempt_users)) {
				PAM_LOG("%s - user \"%s\" is exempt from the "
				    "default quota", __FILE__, session_user);
				quotavar = NULL;
			} else
				quotavar = &tmpquotaconf->default_quota;
		} else
			quotavar = &tmpuserquotaconf->quota;

		if (quotavar != NULL && *quotavar > 0) {
			quota_divisor = 1;
			/* Humanise the quotaval for logging purposes. */
			if (*quotavar > 1073741824) {
				suffix = 'G';
				quota_divisor = 1073741824;
			} else if (*quotavar > 1048576) {
				suffix = 'M';
				quota_divisor = 1048576;
			} else if (*quotavar > 1024) {
				suffix = 'K';
				quota_divisor = 1024;
			}
			sprintf(quotaval, "%jd%c", *quotavar / quota_divisor,
			    suffix);
		}

		PAM_LOG("%s - calling zfs_set_prop(\"%s\", \"%s\", \"%s\")",
		    __FILE__, quotaprop, quotaval, tmpquotaconf->fspath);

		if (zfs_set_prop(quotaprop, quotaval, tmpquotaconf->fspath))
			err++;
	}

	return (err);
}

PAM_EXTERN int
pam_sm_open_session(pam_handle_t *pamh, int flags __unused, int argc __unused,
    const char *argv[] __unused)
{
	const char *cfpath, *dflt_exempt_users;
	int pam_err;

	cfpath = openpam_get_option(pamh, "conf");
	if (cfpath == NULL || strlen(cfpath) == 0) {
		pam_err = PAM_SESSION_ERR;
		goto err;
	}

	dflt_exempt_users = openpam_get_option(pamh, "default_quota_exempt");

	if (openpam_get_option(pamh, "debug"))
		_openpam_debug = 1;

	/* Pull the PAM user into the global session_user variable. */
	pam_err = pam_get_user(pamh, &session_user, NULL);
	if (pam_err != PAM_SUCCESS)
		goto err;

	/* Parse config file. */
	if (parse_config_file(cfpath) != 0) {
		pam_err = PAM_SESSION_ERR;
		goto err;
	}

	if (set_quotas(dflt_exempt_users) != 0)
		pam_err = PAM_SESSION_ERR;

err:
	cleanup();

	if (openpam_get_option(pamh, "no_fail"))
		pam_err = PAM_SUCCESS;

	return (pam_err);
}

PAM_EXTERN int
pam_sm_close_session(pam_handle_t *pamh __unused, int flags __unused,
    int argc __unused, const char *argv[] __unused)
{

	return (PAM_SUCCESS);
}

PAM_MODULE_ENTRY("pam_managequotas");
